/*
Copyright (c) Members of the EGEE Collaboration. 2004.
See http://www.eu-egee.org/partners/ for details on the
copyright holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
 * glite-wms-jdl-adtest.cpp
 * Copyright (c) 2001 The European Datagrid Project - IST programme, all rights reserved.
 * Contributors are mentioned in the code where appropriate.
 */
#include "glite/wmsutils/exception/Exception.h"
#include "glite/wmsutils/jobid/JobId.h"
#include "glite/jdl/extractfiles.h"
#include "glite/jdl/extractfiles.h"
#include "glite/wmsutils/classads/classad_utils.h"
#include "glite/jdl/JDLAttributes.h"
#include "glite-wms-jdl-common.h"

using namespace std ;
using namespace classad;
using namespace glite::jdl ;
using namespace glite::wmsutils::classads;


void testClassAdJdl(const string& jdl){

	const classad::ClassAd *classAd = utilities::parse_classad(jdl);
	string       buffer="";
	PrettyPrint  unp;
	//flat the classAd into one single line
	unp.SetClassAdIndentation(0);
	unp.SetListIndentation(0);
	//Unparse the ClassAd instance
	unp.Unparse(buffer, classAd);
	Ad ad(buffer);
	title (		  "testClassAd reading string    = " +jdl
		+	"\ntestClassAd PARSED classad    = " + buffer);
}


// UNIT TEST
class AdTest : public CppUnit::TestFixture {
	CPPUNIT_TEST_SUITE( AdTest );
	CPPUNIT_TEST( testMerge );
	CPPUNIT_TEST( testClassAd );
	CPPUNIT_TEST( testParseAndReadAd );
	CPPUNIT_TEST( testVarious );
	CPPUNIT_TEST_SUITE_END();
public:




	void testConstructors(){
		Ad ad;
		ad.fromFile(getJdlFile());
		title("FromString "+ad.toString());
	}

	void testMerge() {
		Ad nodes;
		nodes.addAttribute("dependencies", "NodeINVENTEDA");
		nodes.addAttribute("dependencies", "NodeINVENTEDB");

		glite::jdl::Ad jab = createAd();
		glite::jdl::Ad tbm = createAd();
		const string COMMON_ATTR = "common_attribute";
		Ad common;
		common.setAttribute("CmnAd_b",true);
		common.setAttribute("CmnAd_s","FirstAd_VALUE");
		common.setAttribute("CmnAd_i",1);
		common.setAttribute("inherited_jdl", &jab);
		jab.setAttribute(COMMON_ATTR,&common);
		jab.setAttribute("rootAttr01","FirstValue");
		glite::jdl::Ad jab2 = jab;
		common.clear();
		common.setAttribute("CmnAd_s","SecondAd_VALUE");
		common.setAttribute("CmnAd_i",2);
		common.setAttribute("CmnAd_b",false);
		common.setAttribute("fifth","CustomSecondAd_VALUE");
		common.setAttribute("inherited_jdl", &common);
		tbm.setAttribute(COMMON_ATTR,&common);
		if (tbm.hasAttribute("nodes")){
			tbm.delAttribute("nodes");
		}
		tbm.setAttribute("nodes", &nodes);
		tbm.setAttribute("rootAttr01","SecondValue");
		title("Merging two ADS: \n" + jab.toLines() + tbm.toLines());
		jab.merge(tbm);
		// Check
		common.clear();
		common = jab.getAd(COMMON_ATTR);
		CPPUNIT_ASSERT(common.getBool("CmnAd_b")  == false);
		CPPUNIT_ASSERT(common.getString("CmnAd_s") == "SecondAd_VALUE");
		CPPUNIT_ASSERT(common.getInt("CmnAd_i")    == 2);
		title("overwrite MERGED ad is:" + jab.toLines());
		title("Merging two ADS: \n" + jab2.toLines() + tbm.toLines());
		jab2.merge(tbm, false);
		common.clear();
		common = jab2.getAd( COMMON_ATTR );
		CPPUNIT_ASSERT(common.getBool("CmnAd_b")  == true);
		CPPUNIT_ASSERT(common.getString("CmnAd_s") == "FirstAd_VALUE");
		CPPUNIT_ASSERT(common.getInt("CmnAd_i")    == 1);
		title("NOT OVERwritten MERGED ad is:" + jab2.toLines());
	}
	void testClassAd() {
		string
		jdl = "[ attName1 = \"attrValue01\"; attNameZZ = \"file[0]\" ]";
		testClassAdJdl(jdl);

		// jdl = "[ attName1 = \"attrValue01\"; attNameZZ = \"file\[1\]\" ]";  // Warning compile
		// testClassAdJdl(jdl);

		jdl = "[ attName1 = \"attrValue01\"; attNameZZ = \"file\\[2\\]\" ]";
		testClassAdJdl(jdl);

		// jdl = "[ attName1 = \"attrValue01\"; attNameZZ = \"file\\\[3\\\]\" ]"; // Warning compile
		// testClassAdJdl(jdl);

		jdl = "[ attName1 = \"attrValue01\"; attNameZZ = \"file\\\\[4\\\\]\" ]";
		testClassAdJdl(jdl);


	}
	void testParseAndReadAd() {
		glite::jdl::Ad jab = createAd();
		title("CREATED AD: " + jab.toLines() );
	}

	void testVarious(){}



};
// UNIT TEST END
int unitTest(){
	/*  CPP UNIT PART  */
	// Debug mode control:
	setWCDM(false);
	CppUnit::TestResult controller ;
	CppUnit::TestResultCollector result ;
	controller.addListener(&result);
	CppUnit::TestRunner runner ;
	/*   ACTUAL CPP UNIT NAME*/
	runner.addTest (AdTest::suite());

	runner.run (controller);
	CppUnit::TextOutputter outputter (&result , std::cerr);
	outputter.write();
	return   result.wasSuccessful()?0:1;
}

int main(int argc,char *argv[]){
	try{
	if( (argc<2) || (strcmp(argv[1],"--help")==0)|| (strcmp(argv[1],"help")==0)){
		cout << "Usage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
		cout <<"0) Unit Test!!!"<<endl; cout <<" - - - " << endl ;
		cout <<"1) testMerge"<<endl;
		cout <<"2) testClassAd"<<endl;
		cout <<"3) testParseAndReadAd"<<endl;
		cout <<"4) testVarious"<<endl;
		cout <<"5) testConstructors"<<endl;
		return 0;
	}
	if (argc>2){setJdlFile(argv[2]);}
	AdTest test;
	switch (boost::lexical_cast<int>(argv[1])){
		case 0 : unitTest(); break;
		case 1: test.testMerge(); break;
		case 2: test.testClassAd(); break;
		case 3: test.testParseAndReadAd(); break;
		case 4: test.testVarious(); break;
		case 5: test.testConstructors(); break;
		default: break;
	}
	return 0;
	}catch  (glite::wmsutils::exception::Exception &exc){
		cout << "\nOperation not performed\n" ;
		cout <<"Stack...  " << exc.printStackTrace()<<  endl ;
		cout <<"Exc Name  " << exc.getExceptionName()<< endl;
		cout <<"What      " << exc.what()<< endl;
	}catch  (std::exception &exc){
		cout << "\nOperation not performed. std::exception found:\n" ;
		cout <<  exc.what() << endl ;
		cout << "\nUsage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
	}catch  (...){
		cout << "\nOperation not performed (FATAL)...\n" ;
		cout << "\nUsage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
	}
return 1 ;
};

