/*
Copyright (c) Members of the EGEE Collaboration. 2004.
See http://www.eu-egee.org/partners/ for details on the
copyright holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
 * edg-wl-jdl-collect-test.cpp
 * Copyright (c) 2001 The European Datagrid Project - IST programme, all rights reserved.
 * Contributors are mentioned in the code where appropriate.
 */


#include "glite/wmsutils/exception/Exception.h"
#include "glite/jdl/adconverter.h"
#include "glite/jdl/collectionad.h"
#include "glite/jdl/JDLAttributes.h"
#include "glite/jdl/ExpDagAd.h"
#include "glite-wms-jdl-common.h"

using namespace std ;
using namespace glite::jdl ;

int collectServerSide(CollectionAd *collect){
	ExpDagAd *dag = new ExpDagAd(*(AdConverter::collection2dag(collect->toSubmissionString ())));
	title("COLLECTION converted into DagAd:\n"+ dag->toString());
	dagServerSide(*dag);
	return 0;
}

void convertCollectTest() {
	CollectionAd *collect = AdConverter::createCollectionFromPath("clt/*.jdl");
	cout << "Collection converted: " << collect->toLines() << endl ;
	cout << "getNodeStringValueSSS.........." << endl ;
		std::vector< std::pair<std::string,std::vector< std::string > > > values= collect->getNodeStringValues("inputSandbox");
		for (unsigned int i =0 ; i<values.size();i++){
			vector< string > pire=values[i].second;
			cout << "NODE " << values[i].first<< endl ;;
			for (unsigned int j =0 ; j<pire.size();j++){
				cout << j <<") " << pire[j] << endl;
			}
		}
}


void toBCopiedFileList(const std::string &wmp_uri, const std::string &isb_uri, const std::vector <std::string> &paths){
	std::vector <std::pair<std::string, std::string> > to_bcopied ;
	toBcopied("InputSandbox", paths, to_bcopied, wmp_uri, isb_uri);
	for (unsigned int i = 0 ; i< to_bcopied.size() ; i++){
		cout << "SRC:  " << to_bcopied[i].first  <<"  ###   DEST: " << to_bcopied[i].second << endl ;
	}
}




// UNIT TEST
class CollectTest : public CppUnit::TestFixture {
	CPPUNIT_TEST_SUITE( CollectTest );
	CPPUNIT_TEST( testJobClient );
	CPPUNIT_TEST( testJobServer );
	CPPUNIT_TEST_SUITE_END();
public:
	void testJobClient() {
		glite::jdl::CollectionAd jab = createCollectionAd();
		title("Collection: "+jab.toLines());
		title("Submission: "+jab.toSubmissionString());
	}
	void testJobServer() {
		glite::jdl::CollectionAd collect = createCollectionAd();
		title("Collection: "+collect.toLines());
		collectServerSide(&collect);
	}

	void testGetMethods() {
		glite::jdl::CollectionAd collect = createCollectionAd();
		title("Retrieving all ISB nodes from created collection: " + collect.toLines() );
		std::vector< std::pair<std::string,std::vector< std::string > > >
			values= collect.getNodeStringValues("inputSandbox");
			// values= collect.getNodeStringValues(JDL::ISB_BASE_URI);
		string result = "";
		for (unsigned int i =0 ; i<values.size();i++){
			vector< string > pire=values[i].second;
			result += "NODE " + values[i].first +":\n" ;
			for (unsigned int j =0 ; j<pire.size();j++){
				result += "\t" + pire[j] + "\n" ;
			}
		}
		title("Retrieved values: \n" +result);
	}

};
// UNIT TEST END
int unitTest(){
	/*  CPP UNIT PART  */
	// Debug mode control:
	setWCDM(false);
	CppUnit::TestResult controller ;
	CppUnit::TestResultCollector result ;
	controller.addListener(&result);
	CppUnit::TestRunner runner ;
	/*   ACTUAL CPP UNIT NAME*/
	runner.addTest (CollectTest::suite());

	runner.run (controller);
	CppUnit::TextOutputter outputter (&result , std::cerr);
	outputter.write();
	return   result.wasSuccessful()?0:1;
}


int main(int argc,char *argv[]){
	try{
	if( (argc<2) || (strcmp(argv[1],"--help")==0)|| (strcmp(argv[1],"help")==0)){
		cout << "Usage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
		cout <<"0) Unit Test!!!"<<endl; cout <<" - - - " << endl ;
		cout <<"1) testJobClient"<<endl;
		cout <<"2) testJobServer"<<endl;
		cout <<"3) testGetMethods"<<endl;
		return 0;
	}
	if (argc>2){setJdlFile(argv[2]);}
	CollectTest test;
	switch (boost::lexical_cast<int>(argv[1])){
		case 0 : unitTest(); break;
		case 1: test.testJobClient(); break;
		case 2: test.testJobServer(); break;
		case 3: test.testGetMethods(); break;
		default: break;
	}
	return 0;
	}catch  (glite::wmsutils::exception::Exception &exc){
		cout << "\nOperation not performed\n" ;
		cout <<"Stack...  " << exc.printStackTrace()<<  endl ;
		cout <<"Exc Name  " << exc.getExceptionName()<< endl;
		cout <<"What      " << exc.what()<< endl;
	}catch  (std::exception &exc){
		cout << "\nOperation not performed. std::exception found:\n" ;
		cout <<  exc.what() << endl ;
		cout << "\nUsage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
	}
return 1 ;
};

