/*
Copyright (c) Members of the EGEE Collaboration. 2004.
See http://www.eu-egee.org/partners/ for details on the
copyright holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
 * AdConverter Class TEST
 * Copyright (c) 2001 The European Datagrid Project - IST programme, all rights reserved.
 * Contributors are mentioned in the code where appropriate.
 */

#include "glite/wmsutils/exception/Exception.h"
#include "glite/jdl/adconverter.h"
#include "glite/jdl/collectionad.h"
#include "glite/jdl/JDLAttributes.h"
#include "glite/jdl/Ad.h"
#include "glite/jdl/ExpDagAd.h"
#include "glite-wms-jdl-common.h"




#define COLLECTION_PATH "pathcl/*"
#define DAGAD_PATH      "pathdg/*"

using namespace std ;
using namespace glite::jdl ;

/****************************
*   FROM PATH TEST
*****************************/
class  FromPathTest: public CppUnit::TestFixture {
	CPPUNIT_TEST_SUITE( FromPathTest );
	CPPUNIT_TEST( testcreateCollectionFromPath );
	CPPUNIT_TEST( testcreateDagAdFromPath );
	CPPUNIT_TEST_SUITE_END();
public:
	void setUp(){}
	void testcreateCollectionFromPath() {
		string path= getJdlFile();
		CollectionAd *collect;
		collect= AdConverter::createCollectionFromPath(COLLECTION_PATH,DEF_VO);
		ExpDagAd *dagad=AdConverter::collection2dag(collect);
		dagClientSide(*dagad);
		dagServerSide(*dagad);
	}
	void testcreateDagAdFromPath() {
		NodeStruct dependencies;
		ExpDagAd *dagad = AdConverter::createDagAdFromPath(dependencies,DAGAD_PATH,DEF_VO);
		dagClientSide(*dagad);
		dagServerSide(*dagad);
	}
};


/****************************
*   TEMPLATES TEST
*****************************/
class TemplateTest : public CppUnit::TestFixture {
	CPPUNIT_TEST_SUITE( TemplateTest );

	CPPUNIT_TEST(testcreateJobTemplateN);
	CPPUNIT_TEST(testcreateJobTemplateI);
	CPPUNIT_TEST(testcreateJobTemplateM);
	CPPUNIT_TEST(testcreateJobTemplatePt);

	CPPUNIT_TEST(testcreateJobTemplatePm);
	CPPUNIT_TEST(testcreateStringParametricTemplate);

	CPPUNIT_TEST(testcreateDAGTemplate);
	CPPUNIT_TEST(testcreateCollectionTemplate);
	CPPUNIT_TEST_SUITE_END();
public:
	void setUp(){}
	/** JOB TEMPLATES:  */
	void testcreateJobTemplateN(){
		JobAd *jad = AdConverter::createJobTemplate (AdConverter::ADCONV_JOBTYPE_NORMAL,
			DEF_EXECUTABLE,DEF_ARGUMENTS,DEF_REQUIREMENTS, DEF_RANK, DEF_VO);
		jobClientSide(*jad);
		jobServerSide(*jad);
	}
	void testcreateJobTemplateI(){
		JobAd *jad = AdConverter::createJobTemplate (AdConverter::ADCONV_JOBTYPE_INTERACTIVE,
			DEF_EXECUTABLE,DEF_ARGUMENTS, DEF_REQUIREMENTS, DEF_RANK, DEF_VO);
		jobClientSide(*jad);
		jobServerSide(*jad);
	}
	void testcreateJobTemplateM(){
		JobAd *jad = AdConverter::createJobTemplate (AdConverter::ADCONV_JOBTYPE_MPICH,
			DEF_EXECUTABLE,DEF_ARGUMENTS,DEF_REQUIREMENTS, DEF_RANK, DEF_VO);
		jobClientSide(*jad);
		jobServerSide(*jad);
	}

	void testcreateJobTemplatePt(){
		JobAd *jad = AdConverter::createJobTemplate (AdConverter::ADCONV_JOBTYPE_PARTITIONABLE,
			DEF_EXECUTABLE,DEF_ARGUMENTS, DEF_REQUIREMENTS, DEF_RANK, DEF_VO);
		jobClientSide(*jad);
		jobServerSide(*jad);
	}
	/** PARAMETRIC JOB TEMPLATES:  */
	void testcreateJobTemplatePm(){
		JobAd *jad = AdConverter::createJobTemplate (AdConverter::ADCONV_JOBTYPE_PARAMETRIC,
			DEF_EXECUTABLE,DEF_ARGUMENTS, DEF_REQUIREMENTS, DEF_RANK, DEF_VO);
		ExpDagAd* dagad = AdConverter::bulk2dag(jad);
		dagClientSide(*dagad);
	}
	void testcreateStringParametricTemplate(){
		std::vector<std::string> parametrised;
		parametrised.push_back(JDL::ARGUMENTS);
		parametrised.push_back(JDL::STDINPUT);
		std::vector<std::string> parameters;
		parametrised.push_back("first");
		parametrised.push_back("second");
		parametrised.push_back("third");
		JobAd *jad = AdConverter::createStringParametricTemplate (parametrised,parameters,
			DEF_REQUIREMENTS, DEF_RANK, DEF_VO);
		jad->setAttribute(JDL::EXECUTABLE,DEF_EXECUTABLE);
	}
	/** DAG TEMPLATE:  */
	void testcreateDAGTemplate (){
		NodeStruct dependencies, nodeA, nodeB ;
		dependencies.childrenNodes.push_back(new NodeStruct(new string("oneNode")));
		dependencies.childrenNodes.push_back(new NodeStruct(new string("oneOtherNode")));
		ExpDagAd* dagad = AdConverter::createDAGTemplate (dependencies, DEF_REQUIREMENTS, DEF_RANK, DEF_VO);
		dagClientSide(*dagad);
		dagServerSide(*dagad);
	}
	/** COLLECTION TEMPLATE:  */
	void testcreateCollectionTemplate (){
		CollectionAd *collect=AdConverter::createCollectionTemplate(5,DEF_REQUIREMENTS, DEF_RANK, DEF_VO);
		ExpDagAd* dagad = AdConverter::collection2dag(collect);
		dagClientSide(*dagad);
	}
};


int unitTest(){
	/*  CPP UNIT PART  */
	// Debug mode control:
	setWCDM(false);
	CppUnit::TestResult controller ;
	CppUnit::TestResultCollector result ;
	controller.addListener(&result);
	CppUnit::TestRunner runner ;
	runner.addTest (FromPathTest::suite());
	runner.addTest (TemplateTest::suite());
	runner.run (controller);
	CppUnit::TextOutputter outputter (&result , std::cerr);
	outputter.write();
	return   result.wasSuccessful()?0:1;
}

int main(int argc,char *argv[]){
	if( (argc<2) || (strcmp(argv[1],"--help")==0)|| (strcmp(argv[1],"help")==0)){
		cout << "Usage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
		cout <<"0) Unit Test!!!"<<endl;
		cout <<" - - - " << endl ;
		cout <<"1) test.testcreateJobTemplateN()"<<endl;
		cout <<"2) test.testcreateJobTemplateI()"<<endl;
		cout <<"3) test.testcreateJobTemplateM()"<<endl;
		cout <<"4) test.testcreateJobTemplatePt()"<<endl;
		cout <<"5) test.testcreateJobTemplatePm()"<<endl;
		cout <<"6) test.testcreateDAGTemplate()"<<endl;
		cout <<"7) test.testcreateCollectionTemplate()"<<endl;
		cout <<" - - - " << endl ;
		cout <<"8) test.testcreateCollectionFromPath()"<<endl;
		cout <<"9) test.testcreateDagAdFromPath()"<<endl;
		return 0;
	}
	if (argc>2){setJdlFile(argv[2]);}
	try{
			TemplateTest test;
			FromPathTest ftest;
			switch (boost::lexical_cast<int>(argv[1])){
				case 0 : unitTest(); break;
				case 1: test.testcreateJobTemplateN(); break;
				case 2: test.testcreateJobTemplateI(); break;
				case 3: test.testcreateJobTemplateM(); break;
				case 4: test.testcreateJobTemplatePt(); break;
				case 5: test.testcreateJobTemplatePm(); break;
				case 6: test.testcreateDAGTemplate(); break;
				case 7: test.testcreateCollectionTemplate(); break;
				case 8: ftest.testcreateCollectionFromPath(); break;
				case 9: ftest.testcreateDagAdFromPath(); break;
				default: cout << "NO test found..." << endl ;
			}
			return 0 ;
	}catch  (glite::wmsutils::exception::Exception &exc){
		title("\nOperation not performed");
		cout <<"Stack...  " << exc.printStackTrace()<<  endl ;
		cout <<"Exc Name  " << exc.getExceptionName()<< endl;
		cout <<"What      " << exc.what()<< endl;
	}catch  (std::exception &exc){
		cout << "Operation not performed. std::exception found:\n" ;
		cout <<  exc.what() << endl;
		cout << "\nUsage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
	}catch (...){
		title("Uncaught Exception");

	}
};

