/*
Copyright (c) Members of the EGEE Collaboration. 2004.
See http://www.eu-egee.org/partners/ for details on the
copyright holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
 * edg-wl-job-submit.cpp
 * Copyright (c) 2001 The European Datagrid Project - IST programme, all rights reserved.
 * Contributors are mentioned in the code where appropriate.
 */

#include "glite/wmsutils/exception/Exception.h"
#include "glite/wmsutils/jobid/JobId.h"
#include "glite/jdl/JobAd.h"
#include "glite/jdl/JDLAttributes.h"
#include "glite/jdl/ExpDagAd.h"
#include "glite-wms-jdl-common.h"

// DagReplacing Example:
#include "glite/jdl/DAGAd.h"
#include "glite/wmsutils/classads/classad_utils.h"
using namespace classad;
using namespace glite::wmsutils::classads ;

using namespace std ;
using namespace glite::jdl ;

void nodesProva(ExpDagAd &dagadOriginal){
	ExpDagAd dagad= dagadOriginal ;
	title ("nodesProva DBG NODES EXPLORATION get/replace Node approach");
	std::map<std::string,std::string> map = dagad.getJobIdMap();
	std::map<std::string,std::string>::iterator it ;
	NodeAd rif, nodead ;
	for (it= map.begin();it!=map.end();it++){
		glite::wmsutils::jobid::JobId jobid ((*it).first);
		nodead=dagad.getNode(jobid);
		string custom = "Custom";
		// cout << "nodesProva nodead RETRIEVED: " << nodead.toString() << endl ;
		for (unsigned int i = 0 ; i< 10 ; i++){
			nodead.setAttribute (custom,"raskavalue");
			custom+="X";
		}
		dagad.replaceNode(jobid,nodead);
		rif = nodead;
	}
	cout <<"REPLACE UNEXISTING: " << endl ;
	dagad.replaceNode("unexisting", rif);
	// dagad.getNode("unexisting");
	cout <<"Unexisting DONE:"<< dagad.toString()  << endl;

}

// UNIT TEST
class DagAdTest : public CppUnit::TestFixture {
	CPPUNIT_TEST_SUITE( DagAdTest );
	CPPUNIT_TEST( testDagClient );
	CPPUNIT_TEST( testDagServer );
	CPPUNIT_TEST( testJobIdNodes);
	CPPUNIT_TEST_SUITE_END();
public:
	void testDagClient() {
		glite::jdl::ExpDagAd dagad = createDagAd();
		dagClientSide(dagad);
	}
	void testDagServer() {
		glite::jdl::ExpDagAd dagad = createDagAd();
		dagServerSide(dagad);
	}
	void testJobIdNodes() {
		glite::jdl::ExpDagAd dagad = createDagAd();
		insertJobIds(dagad);
		title("JOBIDs Inserted: " +dagad.toString() );
		std::map< std::string, std::string > jmap = dagad.getJobIdMap () ;
		string output = "";
		map< string, string >::iterator it;
		it = jmap.begin();
		for ( ; it!=jmap.end() ; ++it){
			output += "\t" + it->first + " = "+ jmap[it->first] + "\n";
		}
		title ("Map created: NODES/JOBID =\n " + output);
	}
};
int unitTest(){
	/*  CPP UNIT PART  */
	// Debug mode control:
	setWCDM(false);
	CppUnit::TestResult controller ;
	CppUnit::TestResultCollector result ;
	controller.addListener(&result);
	CppUnit::TestRunner runner ;
	/*   ACTUAL CPP UNIT NAME*/
	runner.addTest (DagAdTest::suite());

	runner.run (controller);
	CppUnit::TextOutputter outputter (&result , std::cerr);
	outputter.write();
	return   result.wasSuccessful()?0:1;
}

int main(int argc,char *argv[]){
	try{
		if( (argc<2) || (strcmp(argv[1],"--help")==0)|| (strcmp(argv[1],"help")==0)){
			cout << "Usage : " << argv[0] << " <test number> [<JDL file>]" << endl ;
			cout <<"0) Unit Test!!!"<<endl; cout <<" - - - " << endl ;
			cout <<"1) testDagClient"<<endl;
			cout <<"2) testDagServer"<<endl;
			cout <<"3) testJobIdNodes"<<endl;
			return 0;
		}
		if (argc>2){setJdlFile(argv[2]);}
		DagAdTest test;
		switch (boost::lexical_cast<int>(argv[1])){
			case 0 : unitTest(); break;
			case 1: test.testDagClient(); break;
			case 2: test.testDagServer(); break;
			case 3: test.testJobIdNodes(); break;
			default: break;
		}
		return 0;
	}catch  (glite::wmsutils::exception::Exception &exc){
		cout << "\nOperation not performed\n" ;
		cout <<"Stack...  " << exc.printStackTrace()<<  endl ;
		cout <<"Exc Name  " << exc.getExceptionName()<< endl;
		cout <<"What      " << exc.what()<< endl;
	}catch  (std::exception &exc){
		cout << "\nOperation not performed. std::exception found:\n" ;
		cout <<  exc.what() << endl ;
		cout << "\nUsage : " << argv[0] << " <test number> [<JDL file>]" << endl ;
	}
};

