/*
Copyright (c) Members of the EGEE Collaboration. 2004.
See http://www.eu-egee.org/partners/ for details on the
copyright holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
 * glite-wms-jdl-extracttest.cpp
 * Copyright (c) 2001 The European Datagrid Project - IST programme, all rights reserved.
 * Contributors are mentioned in the code where appropriate.
 */
#include "glite/wmsutils/exception/Exception.h"
#include "glite/wmsutils/jobid/JobId.h"
#include "glite/jdl/JobAd.h"
#include "glite/jdl/NodeAd.h"
#include "glite/jdl/ExpDagAd.h"
#include "glite/jdl/RequestAdExceptions.h"
#include "glite/jdl/extractfiles.h"
#include "glite/wmsutils/classads/classad_utils.h"
#include "glite/jdl/JDLAttributes.h"
#include "glite-wms-jdl-common.h"
#include "glite/jdl/adconverter.h"

#define DEF_PARAM_NUMBER 12


using namespace std ;
using namespace classad;
using namespace glite::jdl ;
using namespace glite::wmsutils::classads ; // classad parser
// UNIT TEST
class ExtractTest : public CppUnit::TestFixture {
	CPPUNIT_TEST_SUITE( ExtractTest );
	CPPUNIT_TEST( testExtractJob );
	CPPUNIT_TEST( testExtractDag );
	CPPUNIT_TEST( testExtractCollection );
	CPPUNIT_TEST( testExtractParametric );
	CPPUNIT_TEST( testExtractGeneric );
	CPPUNIT_TEST( testExtractException );
	CPPUNIT_TEST_SUITE_END();
public:
	void testExtractJob() {
		glite::jdl::JobAd jab = createJobAd();
		title("JobAd created: " + jab.toLines());
		title ("Extracted files...\n" +printExtracted(jab.getExtractedAd()));
	}
	void testExtractDag() {
		glite::jdl::ExpDagAd dagad = createDagAd();
		title("DagAd created: " + dagad.toString());
		title ("Extracted files...\n" +printExtracted(dagad.getExtractedAd()));
	}
	void testExtractCollection() {
		glite::jdl::CollectionAd collect = createCollectionAd();
		title("CollectionAd created: " + collect.toLines());
		title ("Extracted files...\n" +printExtracted(collect.check()->getExtractedAd()));
	}
	void testExtractParametric() {
		glite::jdl::JobAd jab = createJobAd();
		if (!jab.hasAttribute("Parameters")){ jab.setAttribute("Parameters",DEF_PARAM_NUMBER);}
		if (!jab.hasAttribute("JobType","parametric")) {jab.addAttribute("JobType","parametric");}
		title("JobAd Parametric Created:\n" + jab.toLines());
		ExpDagAd *dagad = AdConverter::bulk2dag (&jab);
		title("Parametric Converted into DAG: " +dagad->toString());
		title ("Extracted files...\n" +printExtracted(dagad->getExtractedAd()));
	}
	void testExtractGeneric() {
		glite::jdl::Ad ad = createAd();
		title("Generic Ad created: "+ad.toLines());
		ExtractedAd *extractedAd = new ExtractedAd();
		vector<string> extracted ;
		vector<string> paths = ad.getStringValue(JDL::INPUTSB);
		const string wmpURI  = ad.hasAttribute(JDL::WMPISB_BASE_URI)?ad.getString(JDL::WMPISB_BASE_URI):"";
		const string isbURI  = ad.hasAttribute(JDL::ISB_BASE_URI)?ad.getString(JDL::ISB_BASE_URI):"";
		for (unsigned int i = 0 ; i< paths.size() ; i++){
			extractFiles(JDL::INPUTSB, paths[i], extracted, EXISTENCE , wmpURI, isbURI,extractedAd);
		}
		title ("Extracted files...\n" +printExtracted(extractedAd));
	}
	void testExtractException() {
		ExtractedAd *extractedAd = new ExtractedAd();
		const string UNEXISTING="UNEXISTING_PATH";
		try{
			extractedAd->addFileAd(UNEXISTING);
		}catch (AdSemanticPathException &exc){
			title("AdSemanticPathException rightly caught");
		}
	}
};
// UNIT TEST END
int unitTest(){
	/*  CPP UNIT PART  */
	// Debug mode control:
	setWCDM(false);
	CppUnit::TestResult controller ;
	CppUnit::TestResultCollector result ;
	controller.addListener(&result);
	CppUnit::TestRunner runner ;
	/*   ACTUAL CPP UNIT NAME*/
	runner.addTest (ExtractTest::suite());

	runner.run (controller);
	CppUnit::TextOutputter outputter (&result , std::cerr);
	outputter.write();
	return   result.wasSuccessful()?0:1;
}

int main(int argc,char *argv[]){
	try{
	if( (argc<2) || (strcmp(argv[1],"--help")==0)|| (strcmp(argv[1],"help")==0)){
		cout << "Usage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
		cout <<"0) Unit Test!!!"<<endl; cout <<" - - - " << endl ;
		cout <<"1) testExtractJob"<<endl;
		cout <<"2) testExtractDag"<<endl;
		cout <<"3) testExtractCollection"<<endl;
		cout <<"4) testExtractParametric"<<endl;
		cout <<"5) testExtractGeneric"<<endl;
		cout <<"6) testExtractException"<<endl;
		return 0;
	}
	if (argc>2){setJdlFile(argv[2]);}
	ExtractTest test;
	switch (boost::lexical_cast<int>(argv[1])){
		case 0 : unitTest(); break;
		case 1: test.testExtractJob(); break;
		case 2: test.testExtractDag(); break;
		case 3: test.testExtractCollection(); break;
		case 4: test.testExtractParametric(); break;
		case 5: test.testExtractGeneric(); break;
		case 6: test.testExtractException(); break;
		default: break;
	}
	return 0;
	}catch  (glite::wmsutils::exception::Exception &exc){
		cout << "\nOperation not performed\n" ;
		cout <<"Stack...  " << exc.printStackTrace()<<  endl ;
		cout <<"Exc Name  " << exc.getExceptionName()<< endl;
		cout <<"What      " << exc.what()<< endl;
	}catch  (std::exception &exc){
		cout << "\nOperation not performed. std::exception found:\n" ;
		cout <<  exc.what() << endl ;
		cout << "\nUsage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
	}catch  (...){
		cout << "\nOperation not performed (FATAL)...\n" ;
		cout << "\nUsage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
	}
return 1 ;
};
