/*
Copyright (c) Members of the EGEE Collaboration. 2004.
See http://www.eu-egee.org/partners/ for details on the
copyright holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
 * glite-wms-jdl-jopbadtest.cpp
 * Copyright (c) 2001 The European Datagrid Project - IST programme, all rights reserved.
 * Contributors are mentioned in the code where appropriate.
 */
#include "glite/wmsutils/exception/Exception.h"
#include "glite/wmsutils/jobid/JobId.h"
#include "glite/jdl/JobAd.h"
#include "glite/jdl/NodeAd.h"
#include "glite/jdl/ExpDagAd.h"
//#include "glite/jdl/extractfiles.h"
#include "glite/wmsutils/classads/classad_utils.h"
#include "glite/jdl/JDLAttributes.h"
#include "glite-wms-jdl-common.h"

using namespace std ;
using namespace classad;
using namespace glite::jdl ;
using namespace glite::wmsutils::classads ; // classad parser


bool multiProva(JobAd &jab){
	title("multiProva...");
	vector<string> multiAttributes;
	multiAttributes.push_back("GlueServiceAccessPointURL");
	multiAttributes.push_back("GlueServiceAccessControlRule");
	multiAttributes.push_back("GlueServiceInformationServiceURL");
	multiAttributes.push_back("GlueChunkKey");
	multiAttributes.push_back("GlueSEHostingSL");
	multiAttributes.push_back("GlueSEType");
	multiAttributes.push_back("GlueSEAccessProtocolSupportedSecurity");
	multiAttributes.push_back("GlueSLService");
	multiAttributes.push_back("GlueSLLocalFileSystemClient");
	multiAttributes.push_back("GlueSAAccessControlBaseRule");
	jab.checkMultiAttribute(multiAttributes);
	return false;
}


bool getType(glite::jdl::JobAd *jad){
	title("getType...");
	std::vector< std::string > attributes= jad->attributes ();
	for (unsigned int i =0 ;i< attributes.size(); i++){
		switch( jad->getType(attributes[i])){
			case Ad::TYPE_INTEGER:
			break;
			case Ad::TYPE_BOOLEAN:
			break;
			case Ad::TYPE_STRING:
			break;
			case Ad::TYPE_REAL:
			break;
			case Ad::TYPE_CLASSAD:
			break;
			case Ad::TYPE_EXPRESSION:
			break;
			case Ad::TYPE_UNKNOWN:
			default:
				cout << "Default/Error!" << endl ;
				return true;
		}
	}
	return (attributes.size()==0);
}




bool parseClassad(const std::string &jdl){
	bool result = false;
	try{
		title("parseClassad:\n" +jdl);
		classad::ClassAdParser par ;
		par.ParseExpression(jdl);
	}catch  (std::exception &exc){
			cout << "\nOperation not performed. std::exception found:\n" ;
			cout <<  exc.what() << endl ;
			result = true;
	}

	try{
		title("NOW parse_classad:");
		parse_classad(jdl);
	}catch  (std::exception &exc){
			cout << "\nOperation not performed. std::exception found:\n" ;
			cout <<  exc.what() << endl ;
			result = true;
	}
	try{
		title("Ad creating..");
		Ad ad (jdl);
	}catch  (std::exception &exc){
			cout << "\nOperation not performed. std::exception found:\n" ;
			cout <<  exc.what() << endl ;
			result = true;
	}
	return result;
}



// UNIT TEST
class JobAdTest : public CppUnit::TestFixture {
	CPPUNIT_TEST_SUITE( JobAdTest );
	CPPUNIT_TEST( testJobClient );
	CPPUNIT_TEST( testJobServer );
	CPPUNIT_TEST( testSubmissionString );
	CPPUNIT_TEST( testVarious );
	CPPUNIT_TEST_SUITE_END();
public:
	void testJobClient() {
		glite::jdl::JobAd jab = createJobAd();
		jobClientSide(jab);
	}
	void testJobServer() {
		glite::jdl::JobAd jab = createJobAd();
		jab.check(false);
		jobServerSide(jab);
	}
	void testSubmissionString() {
		glite::jdl::JobAd jab = createJobAd();
		jab.setDefaultRank(DEF_RANK);
		jab.setDefaultReq(DEF_REQUIREMENTS);
		string jdlString = jab.toSubmissionString();
		CPPUNIT_ASSERT(jdlString==jab.toSubmissionString());
		title(jdlString);
	}
	void testVarious() {
		int current_step  = 5 ;
		glite::jdl::JobAd jab = createJobAd();
		JobAd *jad = new JobAd (jab.toString());
		string stringjid = "https://lbproxyServer:76555/uniqueString";
		Ad statead;
		if (jad->hasAttribute(JDL::CHKPT_JOBSTATE)) {
			statead = jad->getAd(JDL::CHKPT_JOBSTATE);
			if (statead.hasAttribute(JDL::CHKPT_STATEID)) {
				statead.delAttribute(JDL::CHKPT_STATEID);
			}
			title("Setting attribute JDL::CHKPT_STATEID");
			statead.addAttribute(JDL::CHKPT_STATEID, stringjid);
		}else {
			statead.setAttribute(JDL::CHKPT_STATEID, stringjid);
			Ad emptyad;
			statead.setAttribute(JDL::CHKPT_DATA, &emptyad);
			if (jad->getType(JDL::CHKPT_STEPS) == Ad::TYPE_INTEGER) {
				statead.setAttribute(JDL::CHKPT_STEPS,
					jad->getIntValue(JDL::CHKPT_STEPS)[0]);
			} else {
				statead.setAttributeExpr(JDL::CHKPT_STEPS,
					jad->lookUp(JDL::CHKPT_STEPS));
			}
			statead.setAttribute(JDL::CHKPT_CURRENTSTEP, current_step);
		}
		title ("testVarious OK: " + jad->toString());
		delete jad;
		title ("testVarious OK: " + statead.toString());
	}


};
// UNIT TEST END
int unitTest(){
	/*  CPP UNIT PART  */
	// Debug mode control:
	setWCDM(false);
	CppUnit::TestResult controller ;
	CppUnit::TestResultCollector result ;
	controller.addListener(&result);
	CppUnit::TestRunner runner ;
	/*   ACTUAL CPP UNIT NAME*/
	runner.addTest (JobAdTest::suite());

	runner.run (controller);
	CppUnit::TextOutputter outputter (&result , std::cerr);
	outputter.write();
	return   result.wasSuccessful()?0:1;
}

int main(int argc,char *argv[]){
	try{
	if( (argc<2) || (strcmp(argv[1],"--help")==0)|| (strcmp(argv[1],"help")==0)){
		cout << "Usage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
		cout <<"0) Unit Test!!!"<<endl; cout <<" - - - " << endl ;
		cout <<"1) testJobClient"<<endl;
		cout <<"2) testJobServer"<<endl;
		cout <<"3) testSubmissionString"<<endl;
		cout <<"4) testVarious"<<endl;
		return 0;
	}
	if (argc>2){setJdlFile(argv[2]);}
	JobAdTest test;
	switch (boost::lexical_cast<int>(argv[1])){
		case 0 : unitTest(); break;
		case 1: test.testJobClient(); break;
		case 2: test.testJobServer(); break;
		case 3: test.testSubmissionString(); break;
		case 4: test.testVarious(); break;
		default: break;
	}
	return 0;
	}catch  (glite::wmsutils::exception::Exception &exc){
		cout << "\nOperation not performed\n" ;
		cout <<"Stack...  " << exc.printStackTrace()<<  endl ;
		cout <<"Exc Name  " << exc.getExceptionName()<< endl;
		cout <<"What      " << exc.what()<< endl;
	}catch  (std::exception &exc){
		cout << "\nOperation not performed. std::exception found:\n" ;
		cout <<  exc.what() << endl ;
		cout << "\nUsage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
	}catch  (...){
		cout << "\nOperation not performed (FATAL)...\n" ;
		cout << "\nUsage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
	}
return 1 ;
};
