/*
Copyright (c) Members of the EGEE Collaboration. 2004.
See http://www.eu-egee.org/partners/ for details on the
copyright holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
 * edg-wl-jdl-collect-test.cpp
 * Copyright (c) 2001 The European Datagrid Project - IST programme, all rights reserved.
 * Contributors are mentioned in the code where appropriate.
 */


#include "glite/wmsutils/exception/Exception.h"
#include "glite/jdl/adconverter.h"
#include "glite/jdl/parametricad.h"
#include "glite/jdl/JDLAttributes.h"
#include "glite/jdl/ExpDagAd.h"
#include "glite/jdl/JDLAttributes.h"
#include "glite-wms-jdl-common.h"

#define DEF_PARAM_NUMBER 12



using namespace std ;
using namespace glite::jdl ;





string replace ( const string& source , const string& s_old ,const string& s_new ) {
	cout << "Replace called...\n"
		<< source << "\n"
		<< s_old <<  "\n"
		<< s_new <<  "\n";
	int strReplaced=0;
	string result = source ;
	while (result.find (s_old )<result.length()){
		cout << "result= " << result ;
		if (strReplaced > 10){
			return "OVERLAPSED STRING";
		}
		result.replace (  result.find (s_old )  , s_old.length() , s_new );
		cout << "result NOW is " << result ;
		strReplaced++;
	}
	return result ;
}




// UNIT TEST
class ParametricTest : public CppUnit::TestFixture {
	CPPUNIT_TEST_SUITE( ParametricTest );
	CPPUNIT_TEST( testParametricClient );
	CPPUNIT_TEST( testParametricString );
	CPPUNIT_TEST( testParametricServer );
	CPPUNIT_TEST( testParametricAd );
	CPPUNIT_TEST( testVarious );
	CPPUNIT_TEST_SUITE_END();
public:
	void testParametricClient() {
		glite::jdl::JobAd jab = createJobAd();
		if (!jab.hasAttribute("Parameters")){ jab.setAttribute("Parameters",DEF_PARAM_NUMBER);}
		if (!jab.hasAttribute("JobType","parametric")) {jab.addAttribute("JobType","parametric");}
		title("JobAd Created:\n" + jab.toLines());
		ExpDagAd *dagad = AdConverter::bulk2dag (&jab,1);
		title("Parametric Converted into DAG: " +dagad->toString());
		dagClientSide(*dagad);
	}
	void testParametricString() {
		glite::jdl::Ad ad = createJobAd();
		string source= ad.toString();
		string s_old ="_PARAM_"  ;
		vector<string> s_news = ad.getStringValue("Parameters");
		for (unsigned int i = 0; i<s_news.size(); i++){
			string result = replace (source, s_old, s_news[i]);
			title("TRANSFORMED STRING= " + result );
		}
	}
	void testParametricServer() {
		glite::jdl::JobAd jab = createJobAd();
		if (!jab.hasAttribute("Parameters")){ jab.setAttribute("Parameters",DEF_PARAM_NUMBER);}
		if (!jab.hasAttribute("JobType","parametric")) {jab.addAttribute("JobType","parametric");}
		title("Parametric Created:\n" + jab.toLines());
		ExpDagAd *dagad = AdConverter::bulk2dag (&jab);
		title("Parametric Converted into DAG: " +dagad->toString());
  		dagServerSide(*dagad);
	}


	void testParametricAd() {
		glite::jdl::Ad parametric = createAd();
	}



	void testVarious() {
		cout << "NOT YET IMPLEMENTED" << endl ;
	}
};
// UNIT TEST END
int unitTest(){
	/*  CPP UNIT PART  */
	// Debug mode control:
	setWCDM(false);
	CppUnit::TestResult controller ;
	CppUnit::TestResultCollector result ;
	controller.addListener(&result);
	CppUnit::TestRunner runner ;
	/*   ACTUAL CPP UNIT NAME*/
	runner.addTest (ParametricTest::suite());

	runner.run (controller);
	CppUnit::TextOutputter outputter (&result , std::cerr);
	outputter.write();
	return   result.wasSuccessful()?0:1;
}

int main(int argc,char *argv[]){
	try{
	if( (argc<2) || (strcmp(argv[1],"--help")==0)|| (strcmp(argv[1],"help")==0)){
		cout << "Usage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
		cout <<"0) Unit Test!!!"<<endl; cout <<" - - - " << endl ;
		cout <<"1) testParametricClient"<<endl;
		cout <<"2) testParametricString"<<endl;
		cout <<"3) testParametricServer"<<endl;
		cout <<"4) testParametricAd"<<endl;
		cout <<"5) testVarious"<<endl;
		return 0;
	}
	if (argc>2){setJdlFile(argv[2]);}
	ParametricTest test;
	switch (boost::lexical_cast<int>(argv[1])){
		case 0 : unitTest(); break;
		case 1: test.testParametricClient(); break;
		case 2: test.testParametricString(); break;
		case 3: test.testParametricServer(); break;
		case 4: test.testVarious(); break;
		default: break;
	}
	return 0;
	}catch  (glite::wmsutils::exception::Exception &exc){
		cout << "\nOperation not performed\n" ;
		cout <<"Stack...  " << exc.printStackTrace()<<  endl ;
		cout <<"Exc Name  " << exc.getExceptionName()<< endl;
		cout <<"What      " << exc.what()<< endl;
	}catch  (std::exception &exc){
		cout << "\nOperation not performed. std::exception found:\n" ;
		cout <<  exc.what() << endl ;
		cout << "\nUsage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
	}catch  (...){
		cout << "\nOperation not performed (FATAL)...\n" ;
		cout << "\nUsage : " << argv[0] << "<test number> [<JDL file>]" << endl ;
	}
return 1 ;
};




